;; ___  ___                 ______                _    
;; |  \/  |                 | ___ \              | |   
;; | .  . | ___   ___  _ __ | |_/ /_ __ ___  __ _| | __
;; | |\/| |/ _ \ / _ \| '_ \| ___ \ '__/ _ \/ _` | |/ /
;; | |  | | (_) | (_) | | | | |_/ / | |  __/ (_| |   < 
;; \_|  |_/\___/ \___/|_| |_\____/|_|  \___|\__,_|_|\_\
;;                                                                                                       
;; Copyright (C) 2020  
;; This file is part of the Amstrad game: MoonBreak
;;
;; Authors:
;; Ricardo Antonio, Ruiz Falcó (@R1ckyman) (ricky.rfr@gmail.com)
;; Alejandro Vicente, Rodríguez Segado (@VicenteSegado) (alejandrors13091993@gmail.com)
;; Enrique Antonio, Cores Rodríguez (@QuiqueCores) (encorod@gmail.com)
;;
;; This program is free software: you can redistribute it and/or modify
;; it under the terms of the GNU General Public License as published by
;; the Free Software Foundation, either version 3 of the License, or
;; (at your option) any later version.
;; 
;; This program is distributed in the hope that it will be useful,
;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;; GNU General Public License for more details.
;; 
;; You should have received a copy of the GNU General Public License
;; along with this program.  If not, see <https://www.gnu.org/licenses/>
.include "cmp/entity.h.s"
.include "cmp/array_structure.h.s"
.include "cpctelera.h.s"

.module Components_Manager

DefineComponentPointerTable _entities, e_cmpID_Num_Components, max_entities
_components_size = . - _entities_components

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;  COMPONENT_MANAGER::init
;;      Fill the array of components with 0's and then create pointers to each
;;      of them according to the number of components.
;;  INPUT:
;;  MODIFY:
;;       A, HL, DE, BC
;;
components_manager_init::
    ;;  Set all values of the pointer array to nullptr
    ;;  filling in the array with 0's.
    ld  hl, #_entities_components
    ld  de, #_entities_components+1
    ld  (hl), #0
    ld  bc, #_components_size-1
    ldir

    ;;  Set values for all Pend pointers in all pointer array.
    ld b, #e_cmpID_Num_Components   ;; Number of components.
_init_next_pend:
    ;; Get in HL a pointer to the next component.
    ld  a, b
    dec a
    call components_manager_getComponentPtrHL
    ld  e, l
    ld  d, h
    inc de
    inc de
    ld  (hl), e
    inc hl
    ld  (hl), d
djnz _init_next_pend

    ret

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;  COMPONENT_MANAGER::getComponentPtrHL
;;      Loads HL with the address of the component
;;      structure of pointers (including the pend pointer).
;;  INPUT:
;;      A:  Component ID
;;  MODIFY:
;;       AF, HL
;;  RETURNS
;;      HL: Pointer to the start of the component structure.
;;
components_manager_getComponentPtrHL::
    ld  hl, #_entities_access_table
    add a
    add_hl_a
    ;;  Get pointer to the component structure.
    ld  a, (hl)
    inc hl
    ld  h, (hl)
    ld  l, a
    ret

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;  COMPONENT_MANAGER::getArrayHL
;;      Loads HL with the address of the array containing entity pointers in the
;;      component structure for component A.
;;  INPUT:
;;  MODIFY:
;;       A, HL
;;
components_manager_getArrayHL::
    call components_manager_getComponentPtrHL
    inc hl
    inc hl
    ret

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;  COMPONENT_MANAGER::add
;;      Adda a pointer to the entity in the corresponding array of components.
;;  INPUT:
;;      IX: Pointer to entity.
;;  MODIFY:
;;       A, HL
;;  RETURNS:
;;
components_manager_add::
    ;;  Gets the pointer to the component's array.
	call components_manager_getComponentPtrHL

    ;;  DE = Pointer to the first free position of the array of pointers.
    ld e, (hl)
    inc hl
    ld d, (hl)
    dec hl

    ;;  Copy the pointer of the entity that handles that component to the
    ;;  array of pointers.
    ld__a_ixl
    ld (de), a
    inc de
    ld__a_ixh
    ld (de), a
    inc de

    ;;  (HL) = Next free position of the pointer array.
    ld (hl), e
    inc hl
    ld (hl), d

	ret

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;  COMPONENT_MANAGER::removePtr
;;      Remove the pointer from the destroyed entity replacing it with the
;;      last pointer of the array.
;;  INPUT:
;;		IX: Entity pointer.
;;		A:  Component ID.
;;  MODIFY:
;;      HL, DE, BC, A
;;  RETURNS:
;;	   
components_manager_removePtr::
    ;;  HL = First pointer of the relevant component array.
	call components_manager_getComponentPtrHL
    push hl
    ;;  DE = Address insert new pointer.
    ld e, (hl)
    inc hl
    ld d, (hl)
    inc hl      ;;  HL = Address to the first pointer of the array.

	__loop:
	ld a, (hl)
	inc hl
    cp__ixl
    jr nz, __loop

	ld a, (hl)
	inc hl
    cp__ixh
    jr nz, __loop

__delete_pointer:
    ;;  The last pointer is copied in the direction of the pointer to be deleted.
    ;;  HL = Pointer direction to remove.
    ;;  DE = Last array pointer.
    dec de
    ld a, (de)
    dec hl
    ld (hl), a

    dec de
    ld a, (de)
    dec hl
    ld (hl), a

__last_array_pointer:
    ;;  HL = Address containing the pointer to the next free address in the array.
    pop hl

    ;;  (HL) = New free position.
    ld (hl), e
    inc hl
    ld (hl), d

    ;;  The content of the las pointer of the array becomes null because it has been
    ;;  copied to the position of the deleted pointer.
    xor a
    ld (de), a
    inc de
    ld (de), a
	ret
