;; ___  ___                 ______                _    
;; |  \/  |                 | ___ \              | |   
;; | .  . | ___   ___  _ __ | |_/ /_ __ ___  __ _| | __
;; | |\/| |/ _ \ / _ \| '_ \| ___ \ '__/ _ \/ _` | |/ /
;; | |  | | (_) | (_) | | | | |_/ / | |  __/ (_| |   < 
;; \_|  |_/\___/ \___/|_| |_\____/|_|  \___|\__,_|_|\_\
;;                                                                                                       
;; Copyright (C) 2020  
;; This file is part of the Amstrad game: MoonBreak
;;
;; Authors:
;; Ricardo Antonio, Ruiz Falcó (@R1ckyman) (ricky.rfr@gmail.com)
;; Alejandro Vicente, Rodríguez Segado (@VicenteSegado) (alejandrors13091993@gmail.com)
;; Enrique Antonio, Cores Rodríguez (@QuiqueCores) (encorod@gmail.com)
;;
;; This program is free software: you can redistribute it and/or modify
;; it under the terms of the GNU General Public License as published by
;; the Free Software Foundation, either version 3 of the License, or
;; (at your option) any later version.
;; 
;; This program is distributed in the hope that it will be useful,
;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;; GNU General Public License for more details.
;; 
;; You should have received a copy of the GNU General Public License
;; along with this program.  If not, see <https://www.gnu.org/licenses/>
;;==============================================================================================================================
;;==============================================================================================================================
;;==============================================================================================================================
;;==============================================================================================================================
;;  WEAPON MANAGER
;;		Manager weapons using a basic memory management system.
;;		Allows to create, delete and update weapons.
;;==============================================================================================================================
;;==============================================================================================================================
;;==============================================================================================================================
;;==============================================================================================================================
.module Weapon_Manager

;;==============================================================================================================================
;;==============================================================================================================================
;;  INCLUDED
;;==============================================================================================================================
;;==============================================================================================================================
;;  CMP/
.include "cmp/weapon.h.s"
.include "cmp/entity.h.s"
.include "cmp/array_structure.h.s"
;;  CPCTELERA
.include "cpctelera.h.s"

;;==============================================================================================================================
;;==============================================================================================================================
;;  WEAPONS ARRAY
;;==============================================================================================================================
;;==============================================================================================================================
DefineBasicStructureArray_Size _weapon, max_weapons, sizeof_w

;;==============================================================================================================================
;;==============================================================================================================================
;;  WEAPON TEMPLATES
;;==============================================================================================================================
;;==============================================================================================================================
rifle:        DefineCmp_Rifle
shotgun:      DefineCmp_Shotgun
enemyShotgun: DefineCmp_EnemyShotgun
enemyRifle:   DefineCmp_EnemyRifle

;;==============================================================================================================================
;;==============================================================================================================================
;;  FUNCTIONS
;;  
;;      INIT:                   Sets the vector of weapons to zero.
;;      CREATE:                 Create a weapon for an entity.
;;      DESTROY:                Destroy the weapons of a dead entity.   (NOT IMPLEMENTED)
;;      GETWEAPONARRAYHL:       Return the array of weapons.
;;      SETPLAYERWEAPONS:       Assigns a rifle and shotgun to the player.
;;      ASSIGNENEMYRIFLE:       Create an enemy rifle for an entity.
;;      ASSIGNENEMYSHOTGUN:     Create an enemy shotgun for an entity.
;;==============================================================================================================================
;;==============================================================================================================================
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;  WEAPON_MANAGER::init
;;      Sets the vector of weapons to zero.
;;  INPUT:
;;  MODIFY:
;;      HL, DE, BC
;;  RETURNS:
;;
weapon_manager_init::
    ld hl, #_weapon_array
    ;;  All reserved memory in the array of weapons is set to 0.
    ld (hl), #0
    ld e, l
    ld d, h
    inc de
    ld bc, #max_weapons * sizeof_w - 1
    ldir
    ret

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;  ENTITY_MANAGER::create
;;      Create a weapon for an entity.
;;  INPUT:
;;      IX: Pointer to entity.
;;      HL: Pointer to the weapon.
;;  MODIFY:
;;       A, HL, BC, DE
;;  RETURNS:
;;
weapon_manager_create::
    push hl
    ld hl, #_weapon_array
    ld bc, #sizeof_w
    
__loop:
    ;;  If the type of the first byte of the address is invalid, the weapon
    ;;  is create at this address.
    ld a, (hl)
    cp #w_type_invalid
    jr nz, __busy_address

__free_address:
    ;;  The offset e_equipped_weapon will contain the pointer to the entity's weapon.
    ld e_equipped_weapon_l(ix), l
    ld e_equipped_weapon_h(ix), h
    ld e, l
    ld d, h
    pop hl
    ldir
    ret

;;  If the address is busy, the following is checked.
__busy_address:
    add hl, bc
    jr __loop

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;  ENTITY_MANAGER::destroy
;;      Destroy the weapons of a dead entity.
;;  INPUT:
;;      IX: Pointer to the entity.
;;  MODIFY:
;;  RETURNS:
;;
weapon_manager_destroy::
    ret

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;  WEAPON_MANAGER::getWeaponArrayHL
;;      Return the array of weapons.
;;  INPUT:
;;  MODIFY:
;;       HL
;;  RETURNS:
;;      HL: Array of weapons.
;;
weapon_manager_getWeaponArrayHL::
    ld hl, #_weapon_array
    ret

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;  WEAPON_MANAGER::setPlayerWeapons
;;      Assigns a rifle and shotgun to the player.
;;  INPUT:
;;  MODIFY:
;;      HL, DE, BC
;;  RETURNS:    
;;
weapon_manager_setPlayerWeapons::
    ;;  Add rifle
    ld hl, #rifle
    call weapon_manager_create

    ;;  Add shotgun
    ld hl, #shotgun
    call weapon_manager_create

    ;;  The player's weapons will be at the beginning of the array.
    ld hl, #_weapon_array
    ld e_equipped_weapon_l(ix), l
    ld e_equipped_weapon_h(ix), h
    ret

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;  WEAPON_MANAGER::assignEnemyRifle
;;      Create an enemy rifle for an entity.
;;  INPUT:
;;  MODIFY:
;;      HL
;;  RETURNS:    
;;
weapon_manager_assignEnemyRifle::
    ld hl, #enemyRifle
    call weapon_manager_create
    ret

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;  WEAPON_MANAGER::assignEnemyShotgun
;;      Create an enemy shotgun for an entity.
;;  INPUT:
;;  MODIFY:
;;      HL
;;  RETURNS:    
;;
weapon_manager_assignEnemyShotgun::
    ld hl, #enemyShotgun
    call weapon_manager_create
    ret
