;; ___  ___                 ______                _    
;; |  \/  |                 | ___ \              | |   
;; | .  . | ___   ___  _ __ | |_/ /_ __ ___  __ _| | __
;; | |\/| |/ _ \ / _ \| '_ \| ___ \ '__/ _ \/ _` | |/ /
;; | |  | | (_) | (_) | | | | |_/ / | |  __/ (_| |   < 
;; \_|  |_/\___/ \___/|_| |_\____/|_|  \___|\__,_|_|\_\
;;                                                                                                       
;; Copyright (C) 2020  
;; This file is part of the Amstrad game: MoonBreak
;;
;; Authors:
;; Ricardo Antonio, Ruiz Falcó (@R1ckyman) (ricky.rfr@gmail.com)
;; Alejandro Vicente, Rodríguez Segado (@VicenteSegado) (alejandrors13091993@gmail.com)
;; Enrique Antonio, Cores Rodríguez (@QuiqueCores) (encorod@gmail.com)
;;
;; This program is free software: you can redistribute it and/or modify
;; it under the terms of the GNU General Public License as published by
;; the Free Software Foundation, either version 3 of the License, or
;; (at your option) any later version.
;; 
;; This program is distributed in the hope that it will be useful,
;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;; GNU General Public License for more details.
;; 
;; You should have received a copy of the GNU General Public License
;; along with this program.  If not, see <https://www.gnu.org/licenses/>
;;==============================================================================================================================
;;==============================================================================================================================
;;==============================================================================================================================
;;==============================================================================================================================
;;  BEHAVIOR_ATTACK
;;      Determines the actions of an entity when in enters attack mode.
;;      These actions will depend on the entity class and the state it is in.
;;      
;;      -BEHAVIOR_ATTACK:            The entity in this state will attack.
;;          -ATTACK_TYPE:            Checks the class of the entity to perform one type off attack or another.
;;              -ATTACK_OR_SCAPE:    If the entity has 1 life, it will flee away from player to heal it.
;;              -MELEE_ATTACK:       The entity changes its state to mobe_to to reach the player.
;;              -SHORT_RANGE_ATTACK: Entity attacks at close range.
;;              -GENERATE:           The entity generates enemies.
;;              -TELEPORT:           Teleports the entity to a random point on the map.
;;      -BHEAVIOR_ESCAPE:            The main task of the entity changes to be healed.
;;      -BEHAVIOR_HEAL:              The entity heals until reaching its maximum life. 
;;==============================================================================================================================
;;==============================================================================================================================
;;==============================================================================================================================
;;==============================================================================================================================
.module Ai_Behavior_Attack

;;  MAN/
.include "man/entity_manager.h.s"
.include "man/game_manager.h.s"
;;  CMP/
.include "cmp/entity.h.s"
.include "cmp/weapon.h.s"
;;  SYS/AI_CONTROL_SYSTEM/
.include "sys/ai_control_system/utilities.h.s"
;;  CPCTELERA/
.include "cpctelera.h.s"

;;==============================================================================================================================
;;==============================================================================================================================
;;  AI_CONTROL_SYSTEM::behavior_attack
;;      Depending on the type of enemy and its attributes makes it does certain
;;      tasks.
;;      If the player is dead the attack is aborted.
;;  INPUT:
;;      IX: Pointer to the entity with AI.
;;  MODIFY:
;;      A
;;  RETURNS:
;;==============================================================================================================================
;;==============================================================================================================================
ai_control_system_behavior_attack::
    call entity_manager_getPlayerPositionIY

    ;;  Check if the player has died.
    ld a, e_type(iy)
    and #e_type_player
    jr z, __aborted_attack

    ;;  Performa ttack based on entity class.
    call ai_control_system_attack_type
    ret

;;  If the player has die, AI continues with it task.
__aborted_attack:
    ld a, e_ai_parent_st(ix)
    ld e_ai_st(ix), a
    ret

;;==============================================================================================================================
;;==============================================================================================================================
;;  AI_CONTROL_SYSTEM::attack_type
;;      Checks the class of the entity to perform one type off attack or another.
;;  INPUT:
;;      IX: Pointer to the entity with AI.
;;  MODIFY:
;;      A
;;  RETURNS:
;;==============================================================================================================================
;;==============================================================================================================================
ai_control_system_attack_type:
    ld a, e_class(ix)
    and #e_class_melee
    call nz, ai_control_system_attack_or_escape

    ld a, e_class(ix)
    and #e_class_generated
    call nz, ai_control_system_melee_attack

    ld a, e_class(ix)
    and #e_class_short_range
    call nz, ai_control_system_melee_attack

    ld a, e_class(ix)
    and #e_class_generator
    call nz, ai_control_system_generate

    ld a, e_class(ix)
    and #e_class_teleporter
    call nz, ai_control_system_teleport
    ret

;;==============================================================================================================================
;;==============================================================================================================================
;;  AI_CONTROL_SYSTEM::attack_or_scape
;;      If the entity has 1 life, it will flee away from player to heal it (SPECIFIC BEHAVIOR FOR MELEE CLASS ENTITIES).
;;  INPUT:
;;      IX: Pointer to the entity with AI.
;;  MODIFY:
;;      A
;;  RETURNS:
;;==============================================================================================================================
;;==============================================================================================================================
ai_control_system_attack_or_escape:
    ;;  If it only has one unit of life, the entity fless.
    ld a, e_att_sta(ix)
    dec a
    jr z, __escape_behavior
    call ai_control_system_melee_attack
    ret
__escape_behavior:
    ld e_ai_st(ix), #e_ai_st_escape
    ret

;;==============================================================================================================================
;;==============================================================================================================================
;;  AI_CONTROL_SYSTEM::melee_attack
;;      The entity changes its state to mobe_to to reach the player.
;;  INPUT:
;;      IX: Pointer to the entity with AI.
;;  MODIFY:
;;      A
;;  RETURNS:
;;==============================================================================================================================
;;==============================================================================================================================
ai_control_system_melee_attack:
    ld a, e_x(iy)
    ld e_ai_aim_x(ix), a

    ld a, e_y(iy)
    ld e_ai_aim_y(ix), a

    ld a, e_ai_st(ix)
    ld e_ai_pre_st(ix), a

    ld e_ai_st(ix), #e_ai_st_move_to
    ret

;;==============================================================================================================================
;;==============================================================================================================================
;;  AI_CONTROL_SYSTEM::short_range_attack
;;      If the player is within shooting range, he will shot,
;;      otherwise he approaches him.
;;  INPUT:
;;      IX: Pointer to the entity with AI.
;;  MODIFY:
;;       A, BC, DE, IY, IX
;;  RETURNS:
;;      A = 0 -> The entity is not within shooting range.
;;      A = 1 -> The entity is a shooting distance.
;;==============================================================================================================================
;;==============================================================================================================================
ai_control_system_short_range_attack::
    ld e, e_equipped_weapon_l(ix)
    ld d, e_equipped_weapon_h(ix)

    ld__iyl_e
    ld__iyh_d

    ld b, w_distance(iy)
    push iy
    call ai_control_system_player_proximity
    ;;  Player coordinates.
    ld e, e_x(iy)
    ld d, e_y(iy)
    pop iy
    jr c, __move_to

    call ai_control_system_adjust_shooting_speeds

    ;;  If the player is within range the entity stops.
    ld e_vx(ix), #0
    ld e_vy(ix), #0

    ld__e_ixl
    ld__d_ixh

    ld__iyl_e
    ld__iyh_d

    push ix
    call game_manager_shot
    pop ix

    ;;  A = 1.
    ;;  The entity is a shooting distance.
    ld a, #1
    ret

__move_to:
    ;;  A = 0.
    ;;  The entity is not within shooting range.
    ld a, #0
    ret

;;==============================================================================================================================
;;==============================================================================================================================
;;  AI_CONTROL_SYSTEM::generate
;;      The entity generates enemies.
;;      When it spawns an enemy it goes to processing state.
;;  INPUT:
;;      IX: Pointer to the entity with AI.
;;  MODIFY:
;;  RETURNS:
;;==============================================================================================================================
;;==============================================================================================================================
ai_control_system_generate::
    call game_manager_generate_enemy
    ld e_ai_pre_st(ix), #e_ai_st_attack
    ld e_ai_st(ix), #e_ai_st_processing
    ret

;;==============================================================================================================================
;;==============================================================================================================================
;;  AI_CONTROL_SYSTEM::teleport
;;      Teleports the entity to a random point on the map.
;;      After teleportation the entity shoots.
;;  INPUT:
;;      IX: Pointer to AI entity.
;;  MODIFY:
;;      DE, IY
;;  RETURNS: 
;;==============================================================================================================================
;;==============================================================================================================================
ai_control_system_teleport::
    call ai_control_system_random_y_position
    call ai_control_system_random_x_position

    ld e_ai_pre_st(ix), #e_ai_st_attack
    ld e_ai_st(ix), #e_ai_st_processing

    ld e, e_equipped_weapon_l(ix)
    ld d, e_equipped_weapon_h(ix)

    ld__iyl_e
    ld__iyh_d
    push iy
    call entity_manager_getPlayerPositionIY

    ld e, e_x(iy)
    ld d, e_y(iy)
    pop iy
    call ai_control_system_adjust_shooting_speeds
    
    ld__e_ixl
    ld__d_ixh

    ld__iyl_e
    ld__iyh_d

    push ix
    call game_manager_shot
    pop ix
    ret

;;==============================================================================================================================
;;==============================================================================================================================
;;  AI_CONTROL_SYSTEM::behavior_escape
;;      The entity changes the state to move to and fleess to the corner.
;;      The main task of the entity changes to be healed.
;;  INPUT:
;;      IX: Pointer to AI entity.
;;  MODIFY:
;;      A
;;==============================================================================================================================
;;==============================================================================================================================
ai_control_system_behavior_escape::
    call ai_control_system_escape_path_search
    ld a, e_ai_st(ix)
    ld e_ai_pre_st(ix), a
    ld e_ai_parent_st(ix), #e_ai_st_heal
    ld e_ai_st(ix), #e_ai_st_move_to
    ret

;;==============================================================================================================================
;;==============================================================================================================================
;;  AI_CONTROL_SYSTEM::behavior_heal
;;      The entity heals until reaching its maximum life.
;;      When he reaches his maximum life he will change his main task to attack. This will not change until the player is killed.
;;  INPUT:
;,      IX: Pointer to AI entity.
;;  MODIFY:
;;      A, B, C
;;==============================================================================================================================
;;==============================================================================================================================
ai_control_system_behavior_heal::
    ld  b, e_ai_st(ix)
    ld  e_ai_pre_st(ix), b
    ;;  Check the stamina of the entity.
    ld  a, e_att_sta(ix)
    ld  c, a
    sub e_att_max_sta(ix)
    jr c, _heal_1_life

;;  If it has healed completely change it main task to attack mode.
_whole_life_healed:
    ld  e_ai_parent_st(ix), #e_ai_st_attack
    ld  e_ai_st(ix), #e_ai_st_processing
    ret

;;  If the stamina is lower than maximun on heals one life.
_heal_1_life:
    ld  a, c
    inc a
    ld  e_att_sta(ix), a
    ld  e_ai_parent_st(ix), b
    ld  e_ai_st(ix), #e_ai_st_processing
    ret
