;; ___  ___                 ______                _    
;; |  \/  |                 | ___ \              | |   
;; | .  . | ___   ___  _ __ | |_/ /_ __ ___  __ _| | __
;; | |\/| |/ _ \ / _ \| '_ \| ___ \ '__/ _ \/ _` | |/ /
;; | |  | | (_) | (_) | | | | |_/ / | |  __/ (_| |   < 
;; \_|  |_/\___/ \___/|_| |_\____/|_|  \___|\__,_|_|\_\
;;                                                                                                       
;; Copyright (C) 2020  
;; This file is part of the Amstrad game: MoonBreak
;;
;; Authors:
;; Ricardo Antonio, Ruiz Falcó (@R1ckyman) (ricky.rfr@gmail.com)
;; Alejandro Vicente, Rodríguez Segado (@VicenteSegado) (alejandrors13091993@gmail.com)
;; Enrique Antonio, Cores Rodríguez (@QuiqueCores) (encorod@gmail.com)
;;
;; This program is free software: you can redistribute it and/or modify
;; it under the terms of the GNU General Public License as published by
;; the Free Software Foundation, either version 3 of the License, or
;; (at your option) any later version.
;; 
;; This program is distributed in the hope that it will be useful,
;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;; GNU General Public License for more details.
;; 
;; You should have received a copy of the GNU General Public License
;; along with this program.  If not, see <https://www.gnu.org/licenses/>
;;==============================================================================================================================
;;==============================================================================================================================
;;==============================================================================================================================
;;==============================================================================================================================
;;  SELECTORS_AND_PARENT_NODES
;;      This class houses the selectors and paren nodes of the AI behaviors.
;;      Each parent node will be composed of a set of behaviors that may vary according to the circumstances of the game.
;;      Whenever a leaf node is reached, the behavior of the entity will return to the parent node, starting the behavior tree
;;      again.
;;      
;;      -PATROL_SELECTOR:       Assign each entity with AI the patrol task.
;;          -BEHAVIOR_PATROL:   Assign the patrol coordinates to each entity with patrol mode.
;;      -GUARD_SELECTOR:        Assign an entity guardian mode.
;;          -BHEAVIOR_GUARD:    The entity remains guarding a perimeter without moving. 
;;==============================================================================================================================
;;==============================================================================================================================
;;==============================================================================================================================
;;==============================================================================================================================
.module Ai_Selectors

;;  MAN/
.include "man/patrol_manager.h.s"
;;  CMP/
.include "cmp/entity.h.s"
;;  SYS/AI_CONTROL_SYSTEM/
.include "sys/ai_control_system/utilities.h.s"
;;  CPCTELERA/

;;==============================================================================================================================
;;==============================================================================================================================
;;  AI_CONTROL_SYSTEM::PATROL_SELECTOR:
;;      Assigns an entity the patrol task.
;;  INPUT:
;;      IX: Pointer to entity with AI.
;;  MODIFY:
;;==============================================================================================================================
;;==============================================================================================================================
ai_control_system_behavior_patrol_selector::
    ld e_ai_parent_st(ix), #e_ai_st_patrol
    ld e_ai_pre_st(ix), #e_ai_st_patrol_selector
    ld e_ai_st(ix), #e_ai_st_patrol
    ret

;;==============================================================================================================================
;;==============================================================================================================================
;;  AI_CONTROL_SYSTEM::BEHAVIOR_PATROL
;;      Sets the path of an entity in patrol mode.
;;      If it suffers any interference, the entity's task could change.
;;  INPUT:
;;      IX: Pointer to entity with AI.
;;  MODIFY:
;;      HL, A
;;==============================================================================================================================
;;==============================================================================================================================
ai_control_system_behavior_patrol::
    ld l, e_ai_patrol_step_l(ix)
    ld h, e_ai_patrol_step_h(ix)

    ld a, (hl)
    cp #patrol_invalid_move_x
    jr z, _reset_patrol

    ld a, (hl)
    ld e_ai_aim_x(ix), a
    inc hl
    ld a, (hl)
    ld e_ai_aim_y(ix), a
    inc hl

    ld e_ai_patrol_step_l(ix), l
    ld e_ai_patrol_step_h(ix), h

    ld e_ai_pre_st(ix), #e_ai_st_patrol
    ld e_ai_st(ix), #e_ai_st_move_to

    ;;  Less aggressive enemy tracking.
    call ai_control_system_interfered
    ret
        
_reset_patrol:
    inc hl
    ld a, (hl)
    inc hl
    ld h, (hl)
    ld e_ai_patrol_step_l(ix), a
    ld e_ai_patrol_step_h(ix), h
    ret

;;==============================================================================================================================
;;==============================================================================================================================
;;  AI_CONTROL_SYSTEM::GUARD_SELECTOR
;;      The entity remains guarding a perimeter without moving. 
;;      If it detects the player, it goes into attack mode and returns to its guard position if it loses sight of it or dies.
;;  INPUT:
;,      IX: Pointer to AI entity.
;;  MODIFY:
;;  RETURNS:
;;==============================================================================================================================
;;==============================================================================================================================
ai_control_system_behavior_guard_selector::
    ld e_ai_parent_st(ix), #e_ai_st_guard
    ld e_ai_pre_st(ix), #e_ai_st_guard_selector
    ld e_ai_st(ix), #e_ai_st_guard
    ret

;;==============================================================================================================================
;;==============================================================================================================================
;;  AI_CONTROL_SYSTEM::BEHAVIOR_GUARD
;;      The entity remains guarding a perimeter without moving. 
;;      If it detects the player, it goes into attack mode and returns to its guard position if it loses sight of it or dies.
;;  INPUT:
;,      IX: Pointer to AI entity.
;;  MODIFY:
;;  RETURNS:
;;==============================================================================================================================
;;==============================================================================================================================
ai_control_system_behavior_guard::
    ;;  Check if any event interferes with guard mode.
    call ai_control_system_interfered

    ;;  If an event occurs, the state of the entity goes to attack.
    ld a, e_ai_st(ix)
    cp #e_ai_st_attack
    ret z

    ;;  If the entity has solved the event, it returns to its guard post.
    ld c, e_initialx(ix)
    ld b, e_initialy(ix)

    ld a, e_x(ix)
    sub c
    jr nz, __return_starting_position
    ld a, e_y(ix)
    sub b
    ret z

    __return_starting_position:
    ld e_ai_aim_x(ix), c
    ld e_ai_aim_y(ix), b

    ld e_ai_st(ix), #e_ai_st_move_to
    ld e_ai_pre_st(ix), #e_ai_st_guard
    ret

