;; ___  ___                 ______                _    
;; |  \/  |                 | ___ \              | |   
;; | .  . | ___   ___  _ __ | |_/ /_ __ ___  __ _| | __
;; | |\/| |/ _ \ / _ \| '_ \| ___ \ '__/ _ \/ _` | |/ /
;; | |  | | (_) | (_) | | | | |_/ / | |  __/ (_| |   < 
;; \_|  |_/\___/ \___/|_| |_\____/|_|  \___|\__,_|_|\_\
;;                                                                                                       
;; Copyright (C) 2020  
;; This file is part of the Amstrad game: MoonBreak
;;
;; Authors:
;; Ricardo Antonio, Ruiz Falcó (@R1ckyman) (ricky.rfr@gmail.com)
;; Alejandro Vicente, Rodríguez Segado (@VicenteSegado) (alejandrors13091993@gmail.com)
;; Enrique Antonio, Cores Rodríguez (@QuiqueCores) (encorod@gmail.com)
;;
;; This program is free software: you can redistribute it and/or modify
;; it under the terms of the GNU General Public License as published by
;; the Free Software Foundation, either version 3 of the License, or
;; (at your option) any later version.
;; 
;; This program is distributed in the hope that it will be useful,
;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;; GNU General Public License for more details.
;; 
;; You should have received a copy of the GNU General Public License
;; along with this program.  If not, see <https://www.gnu.org/licenses/>
;;==============================================================================================================================
;;==============================================================================================================================
;;==============================================================================================================================
;;==============================================================================================================================
;;  UTILITIES
;;      Utility functions that support AI behaviors.
;;      
;;      -INTERFERED:             Change AI tasks according to interference.
;;      -ZONE_SCANNER:           Perform a scan of an area control delimited by the sight value of the entity.
;;      -PLAYER_PROXIMITY:       Detects if the player is within range of the entity.
;;      -ADJUST_SHOOTING_SPEEDS: Adjusts the speed of the enemy weapon based on the player's position.
;;      -BEHAVIOR_PROCESSING:    Puts the entity in processing.
;;      -ESCAPE_PATH_SEARCH:     Find the corner furthest from the player so that the entity escape to there.
;;      -RANDOM_Y_POSITION:      Calculates a random Y-coordinate within the boundaries of the map.
;;      -RANDOM_X_POSITION:      Calculates a random x-coordinate within the boundaries of the map.
;;==============================================================================================================================
;;==============================================================================================================================
;;==============================================================================================================================
;;==============================================================================================================================
.module Ai_Control_System

;;  MAN/
.include "man/entity_manager.h.s"
;;  CMP/
.include "cmp/entity.h.s"
.include "cmp/weapon.h.s"
;;  CPCTELERA/
.include "cpctelera.h.s"
;;  SYS/
.include "sys/physics_system.h.s"

.globl cpct_getRandom_mxor_u8_asm

;;==============================================================================================================================
;;==============================================================================================================================
;;  AI_CONTTROL_SYSTEM::interfered
;;      Change AI tasks according to interference.
;;  INPUT:
;;      IX: Pointer to entity with AI.
;;  MODIFY:
;;  RETURNS:
;;==============================================================================================================================
;;==============================================================================================================================
ai_control_system_interfered::
    call ai_control_system_zone_scanner
    ret c
    ld e_ai_st(ix), #e_ai_st_attack
    ret

;;==============================================================================================================================
;;==============================================================================================================================
;;  AI_CONTTROL_SYSTEM::zone_scanner
;;      Perform a scan of an area control delimited by the sight value of the entity.
;;      If the player is in the sight area, the entity will put into attak mode.
;;      If the player leaves his sight area, he will return to his main task.
;;  INPUT:
;;      IX: Pointer to entity with AI.
;;  MODIFY:
;;      B
;;  RETURNS:
;;      Flag C = 1 -> The player is out of the proximity area determined by the value of register B.
;;      Flag C = 0 -> The player is within of the proximity area determined by the value of register B.
;;==============================================================================================================================
;;==============================================================================================================================
ai_control_system_zone_scanner::
    ;;  B = Sight area.
    ld b, e_ai_sight(ix)
    call ai_control_system_player_proximity
    ret

;;==============================================================================================================================
;;==============================================================================================================================
;;  AI_CONTROL_SYSTEM::player_proximity
;;      Detects if the player is within range of the entity.
;;  INPUT:
;,      IX: Pointer to AI entity.
;;       B: Proximity alarm.
;;  MODIFY:
;;      IY, A
;;  RETURNS:
;;      Flag C = 1 -> The player is out of the proximity area determined by the value of register B.
;;      Flag C = 0 -> The player is within of the proximity area determined by the value of register B.
;;==============================================================================================================================
;;==============================================================================================================================
ai_control_system_player_proximity::
    call entity_manager_getPlayerPositionIY

    ld a, e_endx(ix)
    add b
    sub e_x(iy)
    ret c

    ld a, e_endx(iy)
    add b
    sub e_x(ix)
    ret c

    ld a, e_endy(ix)
    add b
    add b
    sub e_y(iy)
    ret c

    ld a, e_endy(iy)
    add b
    add b
    sub e_y(ix)

    ret

;;==============================================================================================================================
;;==============================================================================================================================
;;  AI_CONTROL_SYSTEM::adjust_shooting_speeds
;;      Adjusts the speed of the enemy weapon based on the player's position.
;;  INPUT:
;;      IX: Pointer to AI entity.
;;       E: X coordinate of the player.
;;       D: Y coordinate of the player.
;;  MODIFY:
;;      IY, A
;;  RETURNS:
;;==============================================================================================================================
;;==============================================================================================================================
ai_control_system_adjust_shooting_speeds::
    ;;  Adjust shooting speed on x axis.
    ld a, e_x(ix)
    sub e
    jr nz, __adjust_speed_x
    ld c, #0
    jr __speed_in_x_axis

    __adjust_speed_x:
    jr c, __positive_x_velocity

    __negative_x_velocity:
        ld c, #-2
        jr __speed_in_x_axis

    __positive_x_velocity:
        ld c, #2

__speed_in_x_axis:
    ld w_vx(iy), c
    ;;  Adjust shooting speed on y axis.
    ld a, e_y(ix)
    sub d
    jr nz, __adjust_speed_y
    ld c, #0
    jr __speed_in_y_axis

    __adjust_speed_y:
    jr c, __positive_y_velocity

    __negative_y_velocity:
        ld c, #-4
        jr __speed_in_y_axis

    __positive_y_velocity:
        ld c, #4

__speed_in_y_axis:
    ld w_vy(iy), c
    ret

;;==============================================================================================================================
;;==============================================================================================================================
;;  AI_CONTROL_SYSTEM::behavior_processing
;;      When the AI performs an action, it says in processing mode
;;      and then returns to the parent node.
;;  INPUT:
;;      IX: Pointer to entity with AI.
;;  MODIFY:
;;      A
;;==============================================================================================================================
;;==============================================================================================================================
ai_control_system_behavior_processing::
    ld a, e_ai_counter(ix)
    dec a
    ld e_ai_counter(ix), a
    ret nz

    ld a, e_ai_processing(ix)
    ld e_ai_counter(ix), a

    ld a, e_ai_parent_st(ix)
    ld e_ai_st(ix), a
    ld e_ai_pre_st(ix), #e_ai_st_processing
    ret

;;==============================================================================================================================
;;==============================================================================================================================
;;  AI_CONTTROL_SYSTEM::escape_patch_search
;;      Find the corner furthest from the player so that the entity escape to there.
;;  INPUT:
;;      IX: Pointer to AI entity.
;;  MODIFY:
;;      IY, A, HL, DE, BC
;;==============================================================================================================================
;;==============================================================================================================================
ai_control_system_escape_path_search::
    ;; IY: Player position
    call entity_manager_getPlayerPositionIY

    ;;  Loads in l screen width.
    ld  a, #screen_width
    sub e_w(ix)
    ld  l, a

    ;;  Loads in h screen height.
    ld  a, #screen_height
    sub e_h(ix)
    ld  h, a

    ;;  Upper left corner.
    ;;  The distance from the player to the upper left corner is calculated.
    ;;  h(n)=|e_x(iy)-0|+|e_y(iy)-0|
    ld  a, e_x(iy)
    add e_y(iy)
    ld  c, a

    ;;  Upper right corner.
    ;;  The distance from the player to the upper right corner is calculated.
    ;;  h(n)=|e_x(iy)-(screen_width-e_w(ix))|+|e_y(iy)-0|
    ld  a, l
    sub e_x(iy)
    add e_y(iy)
    ld  d, a

    ;;  Check which top corner the player is farthest from.
    sub c
    jr  c, _top_left_corner_farthest

;;  The upper right corner is further away from the player.
;;  The calculated value is saved in b and that corner is targer by the AI.
_top_righ_corner_farthest:
    ld e_ai_aim_x(ix), l
    ld e_ai_aim_y(ix), #0
    ld b, d
    jr _calculate_bottom_corners

;;  The upper left corner is further away from the player.
;;  The calculated value is saved in b and that corner is targer by the AI.
_top_left_corner_farthest:
    ld  e_ai_aim_x(ix), #0
    ld  e_ai_aim_y(ix), #0
    ld  b, c

;;  The player's distance from the upper corners is now compared to the lower ones.
_calculate_bottom_corners:
    ;;  Lower left corner.
    ;;  The distance from the player to the lower left corner is calculated.
    ;;  h(n)=|e_x(iy)-0|+|e_y(iy)-(screen_height-e_h(ix))|
    ld  a, h
    sub e_y(iy)
    add e_x(iy)
    ld  c, a

    sub b
    jr  c, _previus_upper_corners_farthest

;;  The lower left corner is further away from the player.
;;  The calculated value is saved in b and that corner is targer by the AI.
_lower_left_corner_farthest:
    ld e_ai_aim_x(ix), #0
    ld e_ai_aim_y(ix), h
    ld  b, c

;;  The previous corners are further away from the player.
_previus_upper_corners_farthest:
    ;;  Lower right corner.
    ;;  The distance from the player to the lower right corner is calculated.
    ;;  h(n)=|e_x(iy)-(screen_width-e_w(ix))|+|e_y(iy)-(screen_height-e_h(ix))|
    ld  a, l
    sub e_x(iy)
    ld  c, a
    ld  a, h
    sub e_y(iy)
    add c

    sub b
    jr  c, _previus_upper_lower_corners_furthest

;;  The lower right corner is further away from the player.
;;  The calculated value is saved in b and that corner is targer by the AI.
_lower_right_corner_farthest:
    ld e_ai_aim_x(ix), l
    ld e_ai_aim_y(ix), h

;;  The previous corners are further away from the player.
_previus_upper_lower_corners_furthest:
    ret

;;==============================================================================================================================
;;==============================================================================================================================
;;  AI_CONTTROL_SYSTEM::random_y_position
;;      Calculates a random y-coordinate within the boundaries of the map.
;;  INPUT:
;;      IX: Pointer to AI entity.
;;  MODIFY:
;;      A, E, B
;;==============================================================================================================================
;;==============================================================================================================================
ai_control_system_random_y_position::
    call cpct_getRandom_mxor_u8_asm

    ld a, #screen_height
    sub e_h(ix)
    ld b, a

	ld a, l
	sub b
	jr nc, __random_y_invalid
    
    __random_y_valid:
        ld a, l
    
    __random_y_invalid:
        ld e_y(ix), a
    ret

;;==============================================================================================================================
;;==============================================================================================================================
;;  AI_CONTTROL_SYSTEM::random_x_position
;;      Calculates a random x-coordinate within the boundaries of the map.
;;  INPUT:
;;      IX: Pointer to AI entity.
;;  MODIFY:
;;      A, E, B
;;==============================================================================================================================
;;==============================================================================================================================
ai_control_system_random_x_position::
    call cpct_getRandom_mxor_u8_asm

    ld a, #screen_width
    sub e_w(ix)
    ld b, a

    ld a, l
    sub b
	jr nc, __random_x_invalid
    
    __random_x_valid:
        jr __assign_random_x_value
    
    __random_x_invalid:
        __loop:
            ld l, a
            sub b
            jr nc, __loop
    
    __assign_random_x_value:
        ld e_x(ix), l
    ret