;; ___  ___                 ______                _    
;; |  \/  |                 | ___ \              | |   
;; | .  . | ___   ___  _ __ | |_/ /_ __ ___  __ _| | __
;; | |\/| |/ _ \ / _ \| '_ \| ___ \ '__/ _ \/ _` | |/ /
;; | |  | | (_) | (_) | | | | |_/ / | |  __/ (_| |   < 
;; \_|  |_/\___/ \___/|_| |_\____/|_|  \___|\__,_|_|\_\
;;                                                                                                       
;; Copyright (C) 2020  
;; This file is part of the Amstrad game: MoonBreak
;;
;; Authors:
;; Ricardo Antonio, Ruiz Falcó (@R1ckyman) (ricky.rfr@gmail.com)
;; Alejandro Vicente, Rodríguez Segado (@VicenteSegado) (alejandrors13091993@gmail.com)
;; Enrique Antonio, Cores Rodríguez (@QuiqueCores) (encorod@gmail.com)
;;
;; This program is free software: you can redistribute it and/or modify
;; it under the terms of the GNU General Public License as published by
;; the Free Software Foundation, either version 3 of the License, or
;; (at your option) any later version.
;; 
;; This program is distributed in the hope that it will be useful,
;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;; GNU General Public License for more details.
;; 
;; You should have received a copy of the GNU General Public License
;; along with this program.  If not, see <https://www.gnu.org/licenses/>
.include "man/components_manager.h.s"
.include "man/game_manager.h.s"
.include "man/entity_manager.h.s"
.include "sys/physics_system.h.s"
.include "cmp/entity.h.s"
.include "cpctelera.h.s"

.module Collision_Detection_System

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;  COLLISION_DETECTION_SYSTEM::init
;;      Initiallizes the collision system by assigning it the array of collideable entity pointers.
;;  INPUT:
;;      IX: Pointer to the entity array
;;       A: Number of elements in the array
;; 
collision_detection_system_init::
    ld a, #e_cmpID_Collisionable
    call components_manager_getArrayHL
    ld  (_ent_array_ptr), hl
    ret

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;  COLLISION_DETECTION_SYSTEM::update
;;      Updates all collisionable entities.
;;  INPUT:
;;      HL: Pointer to collideable entity pointer array.
;;  MODIFY:
;;      HL, DE, A. IX, IY
;;   
collision_detection_system_update::

_ent_array_ptr = . + 1
    ld  hl, #0x0000

_outside_loop:

    ld e, (hl)
    inc hl
    ld d, (hl)
    inc hl

    ld a, e
    or d
    ret z

    push hl

    ld__ixl_e
    ld__ixh_d

_inside_loop:
    ld e, (hl)
    inc hl
    ld d, (hl)
    inc hl

    ld a, e
    or d
    jr z, _inside_loop_end

    ld__iyl_e
    ld__iyh_d
  
    ld a, e_collide_with(ix)
    and e_type(iy)
    jr z, _inside_loop

    call collision_detection_system_check
    jr c, _no_collision_of_entities

_entity_collision:
    call collision_detection_system_behavior

_no_collision_of_entities:
    jr _inside_loop

_inside_loop_end:
    pop hl
    jr _outside_loop

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;  COLLISION_DETECTION_SYSTEM::check
;;      Check if two entities collide.
;;  INPUT:
;;      IX: Entity A
;;      IY: Entity B
;;  MODIFY:
;;      A
;;
collision_detection_system_check::

    ;;  if ( A < B ) _not_collision
    ;;  if ( e_x(ix) + e_w(ix) < e_x(iy) ) __no_collision
    ;;  if ( e_x(ix) + e_w(ix) - e_x(iy) < 0 ) __no_collision
    ld a, e_endx(ix)
    sub e_x(iy)
    ret c

    ;;  if ( C < D ) _not_collision
    ;;  if ( e_x(iy) + e_w(iy) < e_x(ix) ) __no_collision
    ;;  if ( e_x(iy) + e_w(iy) - e_x(ix) < 0 ) __no_collision
    ld a, e_endx(iy)
    sub e_x(ix)
    ret c

    ;;  if ( A < B ) _not_collision
    ;;  if ( e_y(ix) + e_h(ix) < e_y(iy) ) __no_collision
    ;;  if ( e_y(ix) + e_h(ix) - e_y(iy) < 0 ) __no_collision
    ld a, e_endy(ix)
    sub e_y(iy)
    ret c

    ;;  if ( C < D ) _not_collision
    ;;  if ( e_y(iy) + e_h(iy) < e_y(ix) ) __no_collision
    ;;  if ( e_y(iy) + e_h(iy) - e_y(ix) < 0 ) __no_collision
    ld a, e_endy(iy)
    sub e_y(ix)
    ret

collision_detection_system_behavior::
    ld a, e_type(ix)
    cp #e_type_player
    jr nz, _not_collision_player_ix

_collision_player_ix:
    call collision_detection_system_behavior_player
    ret

_not_collision_player_ix:
    ld a, e_type(ix)
    cp #e_type_enemy
    jr nz, _not_collision_enemy_ix

_collision_enemy_ix:
    call collision_detection_system_behavior_enemy
    ret

_not_collision_enemy_ix:
    ld a, e_type(ix)
    and #e_type_shot
    jr nz, _not_collision_shot_ix

_collision_shot_ix:
    call collision_detection_system_behavior_shot

_not_collision_shot_ix:
    ret

;;==============================================================================================================================
;;==============================================================================================================================
;;
;;  COLLISION_DETECTION_SYSTEM::BEHAVIOR_PLAYER
;;      When the player collides with an enemy it is pushed in the opposite direction and decreases its stamina by one unit.
;;      If while being pushed it collides with the edges of the map, it stays in that position and the enemy is pushed back.  
;;  INPUT:
;;      IX: Pointer to player.
;;      IY: Pointer to enemy.
;;  MODIFY:
;;      A, DE, C
;;
;;==============================================================================================================================
;;==============================================================================================================================

collision_detection_system_behavior_player::
    ;;  The collision decreases the player's life and marks him for death if it reaches zero.
    ld a, e_att_sta(ix)
    dec a
    ld e_att_sta(ix), a
    call z, entity_manager_set4destruction

    ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
    ;;  Collision on x-axis:
    ;;      -If the player does not collid with any edge the enemy pushes him on the x axis.
    ;;      -If the player collides with and edge while being pushed, they will stay on the edge and
    ;;      the enemy will be pushed back.
    ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
    ;; Load in a register screen_width-e_w(ix).
    ld a, #screen_width
    sub e_w(ix)
    ld e, a

    ld a, e_x(ix)
    ld c, a
    sub e_x(iy)
    add c
    ld d, a
    cp e
    jr nc, _x_invalid
    ld e_x(ix), a
    jr _end_x_collision

;;  The player has collided with and edge while being pushed.
_x_invalid:
    ;;  The enemy is positioned in the opposite position of the player.
    ld a, e_x(iy)
    ld c, a 
    sub e_x(ix)
    add c
    ld e_x(iy), a
    ;;  It is checked whether is has collided with the left or right edge.
    ld a, d
    add e_w(ix) ;;  The width is added to check if the c flag is activated.
    jr c, _x_invalid_left_edge

;;  The player has collided with the right edge and its position on x axis
;;  becomes screen_width-e_w(ix).
_x_invalid_right_edge:
    ld e_x(ix), e
    jr _end_x_collision

;;  The player has collided with the left edge and its position on x axis
;;  becomes zero.
_x_invalid_left_edge:
    ld e_x(ix), #0

    ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
    ;;  Collision on y-axis:
    ;;      -If the player does not collid with any edge the enemy pushes him on the y axis.
    ;;      -If the player collides with and edge while being pushed, they will stay on the edge and
    ;;      the enemy will be pushed back.
    ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

_end_x_collision:
    ;; Load in a register screen_height-e_h(ix).
    ld a, #screen_height
    sub e_h(ix)
    ld e, a

    ld a, e_y(ix)
    ld c, a
    sub e_y(iy)
    add c
    ld d, a
    cp e
    jr nc, _y_invalid
    ld e_y(ix), a
    ret

;;  The player has collided with and edge while being pushed.
_y_invalid:
    ;;  The enemy is positioned in the opposite position of the player.
    ld a, e_y(iy)
    ld c, a 
    sub e_y(ix)
    add c
    ld e_y(iy), a
    ;;  It is checked whether is has collided with the top or bottom edge.
    ld a, d
    add e_h(ix) ;;  The height is added to check if the c flag is activated.
    jr c, _x_invalid_top_edge

;;  The player has collided with the bottom edge and its position on y axis
;;  becomes screen_height-e_h(ix).
_y_invalid_bottom_edge:
    ld e_y(ix), e
    ret

;;  The player has collided with the top edge and its position on y axis
;;  becomes zero.
_x_invalid_top_edge:
    ld e_y(ix), #0
    ret

;;==============================================================================================================================
;;==============================================================================================================================
;;
;;  COLLISION_DETECTION_SYSTEM::BEHAVIOR_SHOT
;;      ...  
;;  INPUT:
;;      ...
;;  MODIFY:
;;      ...
;;
;;==============================================================================================================================
;;==============================================================================================================================

collision_detection_system_behavior_shot::
    ld a, e_att_sta(ix)
    dec a
    ld e_att_sta(ix), a
    call z, game_manager_entity_destroy
    ld__a_iyl
    ld__ixl_a
    ld__a_iyh
    ld__ixh_a
    ld a, e_att_sta(ix)
    dec a
    ld e_att_sta(ix), a
    call z, entity_manager_set4destruction
    ld e_ai_st(ix), #e_ai_st_attack
    ret

;;==============================================================================================================================
;;==============================================================================================================================
;;
;;  COLLISION_DETECTION_SYSTEM::BEHAVIOR_ENEMY
;;      ...  
;;  INPUT:
;;      ...
;;  MODIFY:
;;      ...
;;
;;==============================================================================================================================
;;==============================================================================================================================

collision_detection_system_behavior_enemy::
    ld a, e_att_sta(ix)
    dec a
    ld e_att_sta(ix), a
    call z, entity_manager_set4destruction
    ld e_ai_st(ix), #e_ai_st_attack
    ld__a_iyl
    ld__ixl_a
    ld__a_iyh
    ld__ixh_a
    ld a, e_att_sta(ix)
    dec a
    ld e_att_sta(ix), a
    call z, game_manager_entity_destroy

    ret
