;; ___  ___                 ______                _    
;; |  \/  |                 | ___ \              | |   
;; | .  . | ___   ___  _ __ | |_/ /_ __ ___  __ _| | __
;; | |\/| |/ _ \ / _ \| '_ \| ___ \ '__/ _ \/ _` | |/ /
;; | |  | | (_) | (_) | | | | |_/ / | |  __/ (_| |   < 
;; \_|  |_/\___/ \___/|_| |_\____/|_|  \___|\__,_|_|\_\
;;                                                                                                       
;; Copyright (C) 2020  
;; This file is part of the Amstrad game: MoonBreak
;;
;; Authors:
;; Ricardo Antonio, Ruiz Falcó (@R1ckyman) (ricky.rfr@gmail.com)
;; Alejandro Vicente, Rodríguez Segado (@VicenteSegado) (alejandrors13091993@gmail.com)
;; Enrique Antonio, Cores Rodríguez (@QuiqueCores) (encorod@gmail.com)
;;
;; This program is free software: you can redistribute it and/or modify
;; it under the terms of the GNU General Public License as published by
;; the Free Software Foundation, either version 3 of the License, or
;; (at your option) any later version.
;; 
;; This program is distributed in the hope that it will be useful,
;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;; GNU General Public License for more details.
;; 
;; You should have received a copy of the GNU General Public License
;; along with this program.  If not, see <https://www.gnu.org/licenses/>
;;==============================================================================================================================
;;==============================================================================================================================
;;==============================================================================================================================
;;==============================================================================================================================
;;  PHYSICS SYSTEMS
;;    MOVES ENTITIES ACCORDING TO THE VELOCITY.
;;    TAKES INTO ACCOUNT SCREEN BORDERS. 
;;==============================================================================================================================
;;==============================================================================================================================
;;==============================================================================================================================
;;==============================================================================================================================

.include "man/entity_manager.h.s"
.include "man/map_manager.h.s"
.include "man/world_manager.h.s"
.include "man/game_manager.h.s"
.include "man/o2_manager.h.s"
.include "sys/physics_system.h.s"
.include "cpctelera.h.s"
.include "cmp/entity.h.s"

.module Physics_System

;;==============================================================================================================================
;;==============================================================================================================================
;;  FUNCTIONS
;;==============================================================================================================================
;;==============================================================================================================================

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;  PHYSICS_SYSTEM::update
;;      Verifies that the entity's pyisics remain within horizontal margins.
;;  INPUT:
;;      IX: Pointer to entity entity.
;;  MODIFY:
;;       A, C
;;
physics_system_check_xaxis::
    ld  a, #screen_width + 1  
    sub e_w(ix)
    ld  c, a                    ;; C = screen width + 1
    ld  a, e_x(ix)              ;; A = Entity.x
    ld e_last_x(ix), a
    add e_vx(ix)                ;; A = Entity.x + Entity.e_vx
    ld (_new_pos_x), a
    cp  c                       ;; Check for Entity.x > Screen width + 1
    ret

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;  PHYSICS_SYSTEM::update
;;      Verifies that the entity's pyisics remain within vertical margins.
;;  INPUT:
;;      IX: Pointer to entity list.
;;  MODIFY:
;;       A, C
;;
physics_system_check_yaxis::
    ld  a, #screen_height + 1  
    sub e_h(ix)
    ld  c, a                    ;; C = screen height + 1
    ld  a, e_y(ix)              ;; A = Entity.y
    ld e_last_y(ix), a
    add e_vy(ix)                ;; A = Entity.y + Entity.e_vy
    ld (_new_pos_y), a
    cp  c                       ;; Check for Entity.y > Screen height + 1
    ret

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;  Calculate the coordinates of a tile, based of the entity coordinates and its speed
;;  INPUT:
;;      - A: Entity coordinate X + its Vx speed
;;  MODIFIES:
;;      - AF, B: Returns coordinate X, C: Returns coordinate Y
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
physics_system_get_tile_coordinates_x::
    ld b, e_x(ix)
    cp b
    jr z, _no_add_width
    jr c, _no_add_width
    ;; x + vx + (width - 1)
    add e_w(ix)
    dec a
_no_add_width:
    ld b, a
    ld a, e_y(ix)
    _check_down_tile = . + 1
    add #0
    ld c, a
    ret
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;  Calculate the coordinates of a tile, based of the entity coordinates and its speed
;;  INPUT:
;;      - A: Entity coordinate Y + its Vy speed
;;  MODIFIES:
;;      - AF, B: Returns coordinate X, C: Returns coordinate Y
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
physics_system_get_tile_coordinates_y::
    ld b, e_y(ix)
    cp b
    jr z, _no_add_height
    jr c, _no_add_height
    ;; y + vy + (height - 1)
    add e_h(ix)
    dec a
_no_add_height:
    ld c, a
    ld a, e_x(ix)
    _check_right_tile = . + 1
    add #0
    ld b, a
    ret
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;  Check if the tile at the given coordinates is solid or not
;;  INPUT:
;;      - B: Coordinate X, C: Coordinate Y
;;  MODIFIES:
;;      - AF, BC, DE, HL
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
physics_system_check_solid_tile::
    call map_manager_get_tile
    ld a, (de)
    ;; First check if it is an special tile
    cp #51
    jr c, _check_solid
    _special_tile:
    cp #59
    jr c, _special_tile_not_solid
    cp #63
    jr nc, _special_tile_not_solid
    ret
    _special_tile_not_solid:
    cp a ;; Set z flag
    ret
    _check_solid:
    or a
    jr nz, _solid_check
    inc a
_solid_check:
    and #0b11111000   ;; The first 7 tiles are not solid, so if a is zero it is not solid, if it is zero it is solid
    ret
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;  Check if the tile at the given coordinates is not solid and is in the edge of the map
;;  INPUT:
;;      - B: Coordinate X, C: Coordinate Y
;;  MODIFIES:
;;      - AF, BC, DE, HL
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
physics_system_check_custom_tile::
    ld a, e_type(ix)
    and #(e_type_player | #e_type_shot)
    ret z
    call map_manager_get_tile
    ld a, (de)
    cp #51
    jr c, _continue
    ld b, a
    ld a, e_type(ix)
    and #e_type_player
    jr nz, _special_player
    _special_bullet:
    call entity_manager_set4destruction
    ret
    _special_player:
    ld a, b
    cp #55
    jr c, _oxigen
    cp #59
    jr c, _door
    _closed_door_or_ship_piece:
    cp #63
    call nc, world_manager_pickup_piece
    ret
    _door:
    call c, world_manager_next_level
    ld a, e_x(ix)
    ld (_new_pos_x), a
    ld a, e_y(ix)
    ld (_new_pos_y), a
    ret
    _oxigen:
    call o2_manager_restore_o2
    ret
    _continue:
    ld b, a
    and #0b11111000   ;; Check if is part of the first 7 tiles, from 5 to 7 will be edge tiles
    ret nz
    ld a, b
    cp #3
    ret c
    ld a, e_type(ix)
    and #e_type_player
    jr nz, _player
    call entity_manager_set4destruction
    ret
    _player:
    ld a, b
    ;; The tile is from [3 to 7]
    sub #3
    jr z, _left
    dec a
    jr z, _right
    dec a
    jr z, _up
_down:
    call world_manager_load_next_map_down
    ld a, #up_position
    ld e_y(ix), a
    ld (_new_pos_y), a
    jr _end_load
_left:
    call world_manager_load_next_map_left
    ld a, #right_position
    ld e_x(ix), a
    ld (_new_pos_x), a
    jr _end_load
_right:
    ld a, #left_position
    ld e_x(ix), a
    ld (_new_pos_x), a
    call world_manager_load_next_map_right
    jr _end_load
_up:
    ld a, #down_position
    ld e_y(ix), a
    ld (_new_pos_y), a
    call world_manager_load_next_map_up
_end_load:
    call world_manager_update_map
    ret

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;  Check the tile where the entity is and processes it
;;  INPUT:
;;      - B: Coordinate X, C: Coordinate Y
;;  MODIFIES:
;;      - AF, BC, DE, HL
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
physics_system_check_tilemap_x::
    push af
    call physics_system_get_tile_coordinates_x
    call physics_system_check_solid_tile
    jr nz, _collision_detected_x
    call physics_system_check_custom_tile
    ld a, e_h(ix)
    dec a
    ld (_check_down_tile), a
    pop af
    call physics_system_get_tile_coordinates_x
    call physics_system_check_solid_tile
    ret
_collision_detected_x:
    pop af
    ret
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;  Check the tile where the entity is and processes it
;;  INPUT:
;;      - B: Coordinate X, C: Coordinate Y
;;  MODIFIES:
;;      - AF, BC, DE, HL
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
physics_system_check_tilemap_y::
    push af
    call physics_system_get_tile_coordinates_y
    call physics_system_check_solid_tile
    jr nz, _collision_detected_y
    call physics_system_check_custom_tile
    ld a, e_w(ix)
    dec a
    ld (_check_right_tile), a
    pop af
    call physics_system_get_tile_coordinates_y
    call physics_system_check_solid_tile
    ret
_collision_detected_y:
    pop af
    ret
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;  PHYSICS_SYSTEM::update
;;      Update the physics of all entities.
;;  INPUT:
;;      IX: Pointer to entity
;;  MODIFY:
;;       A, HL
;;
physics_system_update::
    ld (_end_counter), a

_update_loop:
    ld a, #0
    ld (_check_right_tile), a
    ld (_check_down_tile), a
    ;;ld a, e_vx(ix)    |
    ;;or a              |   Provoca error visual
    ;;jr z, endif_x     |
    call physics_system_check_xaxis
    jr  nc, invalid_x  

valid_x:
    call physics_system_check_tilemap_x
    jr nz, invalid_x
    _new_pos_x = . + 3
    ld  e_x(ix), #0
    add e_w(ix)
    sub #1
    ld e_endx(ix), a
    jr  endif_x

invalid_x:
    ld  a, e_vx(ix)
    neg
    ld  e_vx(ix), a

endif_x:
    ;;ld a, e_vy(ix)    |
    ;;or a              |   Provoca error visual
    ;;jr z, endif_y     |
    call physics_system_check_yaxis
    jr  nc, invalid_y

valid_y:
    call physics_system_check_tilemap_y
    jr nz, invalid_y
    _new_pos_y = . + 3
    ld  e_y(ix), #0
    add e_h(ix)
    sub #1
    ld e_endy(ix), a
    jr  endif_y

invalid_y:
    ld  a, e_vy(ix)
    neg
    ld  e_vy(ix), a

endif_y:

_end_counter = . + 1
	ld a, #0
	dec a
	ret z

	ld (_end_counter), a

    ld e, e_ptr_l(ix)
    ld d, e_ptr_h(ix)
    ld__ixl_e
    ld__ixh_d
    
	jr _update_loop


physics_system_init::
	ret